title 'ユーザー';
description 'ユーザーの定義';

resource user => {
    type => 'object',
    description => 'ユーザー',
    properties => {
        first_name  => {
            type => 'string',
            description => '姓',
            example => '小飼',
        },
        last_name  => {
            type => 'string',
            description => '名',
            example => '弾',
        },
    },
    required => ['first_name', 'last_name'],
};
